@@ -1,6 +1,6 @@
 /* reading patches */
 
-/* $Id: pch.c,v 1.16 1997/05/26 17:52:29 eggert Exp $ */
+/* $Id: pch.c,v 1.26 1997/07/21 17:59:46 eggert Exp $ */
 
 /*
 Copyright 1986, 1987, 1988 Larry Wall
@@ -36,7 +36,10 @@ If not, write to the Free Software Foundation,
 /* Patch (diff listing) abstract type. */
 
 static FILE *pfp;			/* patch file pointer */
-static int p_says_nonexistent[2];	/* [0] for old file, [1] for new */
+static int p_says_nonexistent[2];	/* [0] for old file, [1] for new;
+		value is 0 for nonempty, 1 for empty, 2 for nonexistent */
+static int p_rfc934_nesting;		/* RFC 934 nesting level */
+static time_t p_timestamp[2];		/* timestamps in patch headers */
 static off_t p_filesize;		/* size of the patch file */
 static LINENUM p_first;			/* 1st line number */
 static LINENUM p_newfirst;		/* 1st line number of replacement */
@@ -65,11 +68,11 @@ enum nametype { OLD, NEW, INDEX, NONE };
 static enum diff intuit_diff_type PARAMS ((void));
 static enum nametype best_name PARAMS ((char * const *, int const *));
 static int prefix_components PARAMS ((char *, int));
-static size_t pget_line PARAMS ((int));
+static size_t pget_line PARAMS ((int, int));
 static size_t get_line PARAMS ((void));
 static bool incomplete_line PARAMS ((void));
 static bool grow_hunkmax PARAMS ((void));
-static void malformed PARAMS ((void));
+static void malformed PARAMS ((void)) __attribute__ ((noreturn));
 static void next_intuit_at PARAMS ((file_offset, LINENUM));
 static void skip_to PARAMS ((file_offset, LINENUM));
 
@@ -97,6 +100,7 @@ open_patch_file(filename)
     struct stat st;
     if (!filename || !*filename || strEQ (filename, "-"))
       {
+	file_offset stdin_pos;
 #if HAVE_SETMODE
 	if (binary_transput)
 	  {
@@ -107,10 +111,10 @@ open_patch_file(filename)
 #endif
 	if (fstat (STDIN_FILENO, &st) != 0)
 	  pfatal ("fstat");
-	if (S_ISREG (st.st_mode))
+	if (S_ISREG (st.st_mode) && (stdin_pos = file_tell (stdin)) != -1)
 	  {
 	    pfp = stdin;
-	    file_pos = file_tell (stdin);
+	    file_pos = stdin_pos;
 	  }
 	else
 	  {
@@ -118,7 +122,9 @@ open_patch_file(filename)
 	    pfp = fopen (TMPPATNAME, "w+b");
 	    if (!pfp)
 	      pfatal ("can't create `%s'", TMPPATNAME);
-	    while ((charsread = fread (buf, 1, bufsize, stdin)) != 0)
+	    for (st.st_size = 0;
+		 (charsread = fread (buf, 1, bufsize, stdin)) != 0;
+		 st.st_size += charsread)
 	      if (fwrite (buf, 1, charsread, pfp) != charsread)
 		write_fatal ();
 	    if (ferror (stdin) || fclose (stdin) != 0)
@@ -137,6 +143,8 @@ open_patch_file(filename)
 	  pfatal ("fstat");
       }
     p_filesize = st.st_size;
+    if (p_filesize != (file_offset) p_filesize)
+      fatal ("patch file is too long");
     next_intuit_at (file_pos, (LINENUM) 1);
     set_hunkmax();
 }
@@ -192,10 +200,16 @@ there_is_another_patch()
 	  say (p_base
 	       ? "  Ignoring the trailing garbage.\ndone\n"
 	       : "  I can't seem to find a patch in there anywhere.\n");
+	if (! p_base && p_filesize)
+	  fatal ("Only garbage was found in the patch input.");
 	return FALSE;
     }
     if (skip_rest_of_patch)
+      {
+	Fseek (pfp, p_start, SEEK_SET);
+	p_input_line = p_sline - 1;
 	return TRUE;
+      }
     if (verbosity == VERBOSE)
 	say ("  %sooks like %s to me...\n",
 	    (p_base == 0 ? "L" : "The next patch l"),
@@ -204,21 +218,37 @@ there_is_another_patch()
 	    diff_type == NEW_CONTEXT_DIFF ? "a new-style context diff" :
 	    diff_type == NORMAL_DIFF ? "a normal diff" :
 	    "an ed script" );
-    if (p_indent && verbosity != SILENT)
-	say ("(Patch is indented %d space%s.)\n", p_indent, p_indent==1?"":"s");
+
+    if (verbosity != SILENT)
+      {
+	if (p_indent)
+	  say ("(Patch is indented %d space%s.)\n", p_indent, p_indent==1?"":"s");
+	if (! inname)
+	  {
+	    say ("can't find file to patch at input line %ld\n",
+		 p_sline);
+	    say (strippath == INT_MAX
+	         ? "Perhaps you should have used the -p or --strip option?\n"
+		 : "Perhaps you used the wrong -p or --strip option?\n");
+	  }
+      }
+
     skip_to(p_start,p_sline);
     while (!inname) {
 	if (force || batch) {
-	    say ("No file to patch.  Skipping...\n");
+	    say ("No file to patch.  Skipping patch.\n");
 	    skip_rest_of_patch = TRUE;
 	    return TRUE;
 	}
 	ask ("File to patch: ");
-	inname = fetchname (buf, 0, (int *) 0);
+	inname = fetchname (buf, 0, (time_t *) 0);
 	if (inname)
 	  {
 	    if (stat (inname, &instat) == 0)
-	      inerrno = 0;
+	      {
+		inerrno = 0;
+		invc = -1;
+	      }
 	    else
 	      {
 		perror (inname);
@@ -230,7 +260,7 @@ there_is_another_patch()
 	    ask ("Skip this patch? [y] ");
 	    if (*buf != 'n') {
 		if (verbosity != SILENT)
-		    say ("Skipping patch...\n");
+		    say ("Skipping patch.\n");
 		skip_rest_of_patch = TRUE;
 		return TRUE;
 	    }
@@ -259,11 +289,15 @@ intuit_diff_type()
     char *name[3];
     struct stat st[3];
     int stat_errno[3];
+    int version_controlled[3];
     register enum diff retval;
-    int head_says_nonexistent[2];
 
     name[OLD] = name[NEW] = name[INDEX] = 0;
-    head_says_nonexistent[OLD] = head_says_nonexistent[NEW] = 0;
+    version_controlled[OLD] = -1;
+    version_controlled[NEW] = -1;
+    version_controlled[INDEX] = -1;
+    p_rfc934_nesting = 0;
+    p_timestamp[OLD] = p_timestamp[NEW] = (time_t) -1;
     p_says_nonexistent[OLD] = p_says_nonexistent[NEW] = 0;
     Fseek (pfp, p_base, SEEK_SET);
     p_input_line = p_bline - 1;
@@ -273,7 +307,7 @@ intuit_diff_type()
 	stars_last_line = stars_this_line;
 	this_line = file_tell (pfp);
 	indent = 0;
-	if (! pget_line (0)) {
+	if (! pget_line (0, 0)) {
 	    if (first_command_line >= 0) {
 					/* nothing but deletes!? */
 		p_start = first_command_line;
@@ -284,8 +318,7 @@ intuit_diff_type()
 	    else {
 		p_start = this_line;
 		p_sline = p_input_line;
-		retval = NO_DIFF;
-		goto scan_exit;
+		return NO_DIFF;
 	    }
 	}
 	for (s = buf; *s == ' ' || *s == '\t' || *s == 'X'; s++) {
@@ -304,14 +337,12 @@ intuit_diff_type()
 	    p_indent = indent;		/* assume this for now */
 	}
 	if (!stars_last_line && strnEQ(s, "*** ", 4))
-	    name[OLD] = fetchname (s+4, strippath, &head_says_nonexistent[OLD]);
-	else if (strnEQ(s, "--- ", 4))
-	    name[NEW] = fetchname (s+4, strippath, &head_says_nonexistent[NEW]);
+	    name[OLD] = fetchname (s+4, strippath, &p_timestamp[OLD]);
 	else if (strnEQ(s, "+++ ", 4))
 	    /* Swap with NEW below.  */
-	    name[OLD] = fetchname (s+4, strippath, &head_says_nonexistent[OLD]);
+	    name[OLD] = fetchname (s+4, strippath, &p_timestamp[OLD]);
 	else if (strnEQ(s, "Index:", 6))
-	    name[INDEX] = fetchname (s+6, strippath, (int *) 0);
+	    name[INDEX] = fetchname (s+6, strippath, (time_t *) 0);
 	else if (strnEQ(s, "Prereq:", 7)) {
 	    for (t = s + 7;  ISSPACE ((unsigned char) *t);  t++)
 	      continue;
@@ -326,7 +357,21 @@ intuit_diff_type()
 		revision = savestr (revision);
 		*t = oldc;
 	    }
-	}
+	} else
+	  {
+	    for (t = s;  t[0] == '-' && t[1] == ' ';  t += 2)
+	      continue;
+	    if (strnEQ(t, "--- ", 4))
+	      {
+		time_t timestamp = (time_t) -1;
+		name[NEW] = fetchname (t+4, strippath, &timestamp);
+		if (timestamp != (time_t) -1)
+		  {
+		    p_timestamp[NEW] = timestamp;
+		    p_rfc934_nesting = (t - s) >> 1;
+		  }
+	      }
+	  }
 	if ((diff_type == NO_DIFF || diff_type == ED_DIFF) &&
 	  first_command_line >= 0 &&
 	  strEQ(s, ".\n") ) {
@@ -338,24 +383,32 @@ intuit_diff_type()
 	}
 	if ((diff_type == NO_DIFF || diff_type == UNI_DIFF)
 	    && strnEQ(s, "@@ -", 4)) {
-	    s += 4;
-	    /* `name' and `head_says_nonexistent' are backwards.
-	       Swap the former, and interpret the latter backwards.  */
+
+	    /* `name' and `p_timestamp' are backwards; swap them.  */
+	    time_t ti = p_timestamp[OLD];
+	    p_timestamp[OLD] = p_timestamp[NEW];
+	    p_timestamp[NEW] = ti;
 	    t = name[OLD];
 	    name[OLD] = name[NEW];
 	    name[NEW] = t;
-	    if (head_says_nonexistent[NEW] && ! atol (s))
-	      p_says_nonexistent[OLD] = 1;
+
+	    s += 4;
+	    if (! atol (s))
+	      p_says_nonexistent[OLD] = 1 + ! p_timestamp[OLD];
 	    while (*s != ' ' && *s != '\n')
 	      s++;
 	    while (*s == ' ')
 	      s++;
-	    if (head_says_nonexistent[OLD] && ! atol (s))
-	      p_says_nonexistent[NEW] = 1;
+	    if (! atol (s))
+	      p_says_nonexistent[NEW] = 1 + ! p_timestamp[NEW];
 	    p_indent = indent;
 	    p_start = this_line;
 	    p_sline = p_input_line;
 	    retval = UNI_DIFF;
+	    if (! ((name[OLD] || ! p_timestamp[OLD])
+		   && (name[NEW] || ! p_timestamp[NEW])))
+	      say ("missing header for unified diff at line %ld of patch\n",
+		   p_sline);
 	    goto scan_exit;
 	}
 	stars_this_line = strnEQ(s, "********", 8);
@@ -364,8 +417,8 @@ intuit_diff_type()
 	     || diff_type == NEW_CONTEXT_DIFF)
 	    && stars_last_line && strnEQ (s, "*** ", 4)) {
 	    s += 4;
-	    if (head_says_nonexistent[OLD] && ! atol (s))
-	      p_says_nonexistent[OLD] = 1;
+	    if (! atol (s))
+	      p_says_nonexistent[OLD] = 1 + ! p_timestamp[OLD];
 	    /* if this is a new context diff the character just before */
 	    /* the newline is a '*'. */
 	    while (*s != '\n')
@@ -375,20 +428,23 @@ intuit_diff_type()
 	    p_sline = p_input_line - 1;
 	    retval = (*(s-1) == '*' ? NEW_CONTEXT_DIFF : CONTEXT_DIFF);
 
-	    if (head_says_nonexistent[NEW])
-	      {
-		/* Scan the first hunk to see whether the file appears to
-		   have been deleted.  */
-		file_offset saved_p_base = p_base;
-		LINENUM saved_p_bline = p_bline;
-		p_input_line = p_sline;
-		Fseek (pfp, previous_line, SEEK_SET);
-		if (another_hunk (retval, 0)
-		    && ! p_repl_lines && p_newfirst == 1)
-		  p_says_nonexistent[NEW] = 1;
-		next_intuit_at (saved_p_base, saved_p_bline);
-	      }
+	    {
+	      /* Scan the first hunk to see whether the file contents
+		 appear to have been deleted.  */
+	      file_offset saved_p_base = p_base;
+	      LINENUM saved_p_bline = p_bline;
+	      Fseek (pfp, previous_line, SEEK_SET);
+	      p_input_line -= 2;
+	      if (another_hunk (retval, 0)
+		  && ! p_repl_lines && p_newfirst == 1)
+		p_says_nonexistent[NEW] = 1 + ! p_timestamp[NEW];
+	      next_intuit_at (saved_p_base, saved_p_bline);
+	    }
 
+	    if (! ((name[OLD] || ! p_timestamp[OLD])
+		   && (name[NEW] || ! p_timestamp[NEW])))
+	      say ("missing header for context diff at line %ld of patch\n",
+		   p_sline);
 	    goto scan_exit;
 	}
 	if ((diff_type == NO_DIFF || diff_type == NORMAL_DIFF) &&
@@ -409,12 +465,18 @@ intuit_diff_type()
        (with some modifications if posixly_correct is zero):
 
        - Take the old and new names from the context header if present,
-	 and take the index name from the `Index:' line if present.
+	 and take the index name from the `Index:' line if present and
+	 if either the old and new names are both absent
+	 or posixly_correct is nonzero.
 	 Consider the file names to be in the order (old, new, index).
        - If some named files exist, use the first one if posixly_correct
 	 is nonzero, the best one otherwise.
-       - If no named files exist, some names are given, posixly_correct is
-	 zero, and the patch appears to create a file, then use the best name
+       - If patch_get is nonzero, and no named files exist,
+	 but an RCS or SCCS master file exists,
+	 use the first named file with an RCS or SCCS master.
+       - If no named files exist, no RCS or SCCS master was found,
+	 some names are given, posixly_correct is zero,
+	 and the patch appears to create a file, then use the best name
 	 requiring the creation of the fewest directories.
        - Otherwise, report failure by setting `inname' to 0;
 	 this causes our invoker to ask the user for a file name.  */
@@ -425,6 +487,12 @@ intuit_diff_type()
       {
 	enum nametype i0 = NONE;
 
+	if (! posixly_correct && (name[OLD] || name[NEW]) && name[INDEX])
+	  {
+	    free (name[INDEX]);
+	    name[INDEX] = 0;
+	  }
+
 	for (i = OLD;  i <= INDEX;  i++)
 	  if (name[i])
 	    {
@@ -450,15 +518,57 @@ intuit_diff_type()
 	  {
 	    i = best_name (name, stat_errno);
 
-	    if (p_says_nonexistent[reverse ^ (i == NONE)])
+	    if (i == NONE && patch_get)
+	      {
+		enum nametype nope = NONE;
+
+		for (i = OLD;  i <= INDEX;  i++)
+		  if (name[i])
+		    {
+		      char const *cs;
+		      char *getbuf;
+		      char *diffbuf;
+		      int readonly = outfile && strcmp (outfile, name[i]) != 0;
+
+		      if (nope == NONE || strcmp (name[nope], name[i]) != 0)
+			{
+			  cs = (version_controller
+			        (name[i], readonly, (struct stat *) 0,
+				 &getbuf, &diffbuf));
+			  version_controlled[i] = !! cs;
+			  if (cs)
+			    {
+			      if (version_get (name[i], cs, 0, readonly,
+					       getbuf, &st[i]))
+				stat_errno[i] = 0;
+			      else
+				version_controlled[i] = 0;
+
+			      free (getbuf);
+			      free (diffbuf);
+
+			      if (! stat_errno[i])
+				break;
+			    }
+			}
+
+		      nope = i;
+		    }
+	      }
+
+	    if (p_says_nonexistent[reverse ^ (i == NONE || st[i].st_size == 0)])
 	      {
 		assert (i0 != NONE);
 		if (ok_to_reverse
 		    ("The next patch%s would %s the file `%s',\nwhich %s!",
 		     reverse ? ", when reversed," : "",
-		     i == NONE ? "delete" : "create",
-		     name[i == NONE ? i0 : i],
-		     i == NONE ? "does not exist" : "already exists"))
+		     (i == NONE ? "delete"
+		      : st[i].st_size == 0 ? "empty out"
+		      : "create"),
+		     name[i == NONE || st[i].st_size == 0 ? i0 : i],
+		     (i == NONE ? "does not exist"
+		      : st[i].st_size == 0 ? "is already empty"
+		      : "already exists")))
 		  reverse ^= 1;
 	      }
 
@@ -493,6 +603,7 @@ intuit_diff_type()
 	inname = name[i];
 	name[i] = 0;
 	inerrno = stat_errno[i];
+	invc = version_controlled[i];
 	instat = st[i];
       }
 
@@ -1335,19 +1446,21 @@ another_hunk (difftype, rev)
 static size_t
 get_line ()
 {
-   return pget_line (p_indent);
+   return pget_line (p_indent, p_rfc934_nesting);
 }
 
 /* Input a line from the patch file, worrying about indentation.
    Strip up to INDENT characters' worth of leading indentation.
+   Then remove up to RFC934_NESTING instances of leading "- ".
    Ignore any partial lines at end of input, but warn about them.
    Succeed if a line was read; it is terminated by "\n\0" for convenience.
    Return the number of characters read, including '\n' but not '\0'.
    Return -1 if we ran out of memory.  */
 
 static size_t
-pget_line (indent)
+pget_line (indent, rfc934_nesting)
      int indent;
+     int rfc934_nesting;
 {
   register FILE *fp = pfp;
   register int c;
@@ -1376,6 +1489,23 @@ pget_line (indent)
 
   i = 0;
   b = buf;
+
+  while (c == '-' && 0 <= --rfc934_nesting)
+    {
+      c = getc (fp);
+      if (c == EOF)
+	goto patch_ends_in_middle_of_line;
+      if (c != ' ')
+	{
+	  i = 1;
+	  b[0] = '-';
+	  break;
+	}
+      c = getc (fp);
+      if (c == EOF)
+	goto patch_ends_in_middle_of_line;
+    }
+
   s = bufsize;
 
   for (;;)
@@ -1398,17 +1528,18 @@ pget_line (indent)
 	break;
       c = getc (fp);
       if (c == EOF)
-	{
-	  if (ferror (fp))
-	    read_fatal ();
-	  say ("patch unexpectedly ends in middle of line\n");
-	  return 0;
-	}
+	goto patch_ends_in_middle_of_line;
     }
 
   b[i] = '\0';
   p_input_line++;
   return i;
+
+ patch_ends_in_middle_of_line:
+  if (ferror (fp))
+    read_fatal ();
+  say ("patch unexpectedly ends in middle of line\n");
+  return 0;
 }
 
 static bool
@@ -1533,7 +1664,8 @@ pch_swap()
     return TRUE;
 }
 
-/* Return whether file WHICH (0 = old, 1 = new) appears to be nonexistent.  */
+/* Return whether file WHICH (0 = old, 1 = new) appears to nonexistent.
+   Return 1 for empty, 2 for nonexistent.  */
 
 bool
 pch_says_nonexistent (which)
@@ -1542,6 +1674,16 @@ pch_says_nonexistent (which)
   return p_says_nonexistent[which];
 }
 
+/* Return timestamp of patch header for file WHICH (0 = old, 1 = new),
+   or -1 if there was no timestamp or an error in the timestamp.  */
+
+time_t
+pch_timestamp (which)
+     int which;
+{
+  return p_timestamp[which];
+}
+
 /* Return the specified line position in the old file of the old context. */
 
 LINENUM
@@ -1665,6 +1807,7 @@ do_ed_script (ofp)
 	copy_file (inname, TMPOUTNAME, instat.st_mode);
 	sprintf (buf, "%s %s%s", ed_program, verbosity == VERBOSE ? "" : "- ",
 		 TMPOUTNAME);
+	fflush (stdout);
 	pipefp = popen(buf, binary_transput ? "wb" : "w");
 	if (!pipefp)
 	  pfatal ("can't open pipe to `%s'", buf);
